"""Heating meter module to measure gas consumption, costs and co2 emission. """

# clean
from dataclasses import dataclass
from typing import List, Optional, Any

import pandas as pd
from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import dynamic_component
from hisim import loadtypes as lt
from hisim.component import ComponentInput, OpexCostDataClass
from hisim.components.configuration import EmissionFactorsAndCostsForFuelsConfig
from hisim.dynamic_component import (
    DynamicComponent,
    DynamicConnectionInput,
    DynamicConnectionOutput,
    DynamicComponentConnection,
)
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiTagEnumClass, KpiHelperClass

__authors__ = "Jonas Hoppe"
__copyright__ = ""
__credits__ = ["Jonas Hoppe"]
__license__ = "-"
__version__ = ""
__maintainer__ = ""
__status__ = ""


@dataclass_json
@dataclass
class HeatingMeterConfig(cp.ConfigBase):
    """Heating Meter Config."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return HeatingMeter.get_full_classname()

    building_name: str
    name: str

    @classmethod
    def get_heating_meter_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default GasMeter."""
        return HeatingMeterConfig(
            building_name=building_name,
            name="HeatingMeter",
        )


class HeatingMeter(DynamicComponent):
    """Heating meter class."""

    # Outputs
    HeatAvailable = "HeatAvailable"
    HeatConsumption = "HeatConsumption"
    HeatProduction = "HeatProduction"
    CumulativeConsumption = "CumulativeConsumption"
    CumulativeProduction = "CumulativeProduction"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatingMeterConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(display_in_webtool=True),
    ):
        """Initialize the component."""
        self.grid_energy_balancer_config = config
        self.name = self.grid_energy_balancer_config.name
        self.my_component_inputs: List[DynamicConnectionInput] = []
        self.my_component_outputs: List[DynamicConnectionOutput] = []
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_component_inputs=self.my_component_inputs,
            my_component_outputs=self.my_component_outputs,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.production_inputs: List[ComponentInput] = []
        self.consumption_uncontrolled_inputs: List[ComponentInput] = []

        self.seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep
        # Component has states
        self.state = HeatingMeterState(cumulative_production_in_watt_hour=0, cumulative_consumption_in_watt_hour=0)
        self.previous_state = self.state.self_copy()

        # Outputs
        self.heat_available_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.HeatAvailable,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.HeatAvailable} will follow.",
        )

        self.heat_consumption_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.HeatConsumption,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.HeatConsumption} will follow.",
        )

        self.heat_production_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.HeatProduction,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.HeatProduction} will follow.",
        )

        self.cumulative_heat_consumption_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.CumulativeConsumption,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.CumulativeConsumption} will follow.",
        )

        self.cumulative_heat_production_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.CumulativeProduction,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.CumulativeProduction} will follow.",
        )

        self.add_dynamic_default_connections(self.get_default_connections_from_heat_distribution_system())
        # self.add_dynamic_default_connections(self.get_default_connections_from_simple_dhw_storage())
        self.add_dynamic_default_connections(self.get_default_connections_from_more_advanced_heat_pump())
        self.add_dynamic_default_connections(self.get_default_connections_from_generic_heat_source())

    def get_default_connections_from_heat_distribution_system(
        self,
    ):
        """Get gas heater default connections."""

        from hisim.components.heat_distribution_system import (  # pylint: disable=import-outside-toplevel
            HeatDistribution,
        )

        dynamic_connections = []
        heat_distribution_class_name = HeatDistribution.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=HeatDistribution,
                source_class_name=heat_distribution_class_name,
                source_component_field_name=HeatDistribution.ThermalPowerDelivered,
                source_load_type=lt.LoadTypes.HEATING,
                source_unit=lt.Units.WATT,
                source_tags=[lt.InandOutputType.HEAT_CONSUMPTION],
                source_weight=999,
            )
        )
        return dynamic_connections

    # def get_default_connections_from_simple_dhw_storage(
    #     self,
    # ):
    #     """Get gas heater default connections."""
    #
    #     from hisim.components.simple_dhw_storage import (  # pylint: disable=import-outside-toplevel
    #         SimpleDHWStorage,
    #     )
    #
    #     dynamic_connections = []
    #     dhw_storage_class_name = SimpleDHWStorage.get_classname()
    #     dynamic_connections.append(
    #         dynamic_component.DynamicComponentConnection(
    #             source_component_class=SimpleDHWStorage,
    #             source_class_name=dhw_storage_class_name,
    #             source_component_field_name=SimpleDHWStorage.ThermalPowerConsumptionDHW,
    #             source_load_type=lt.LoadTypes.HEATING,
    #             source_unit=lt.Units.WATT,
    #             source_tags=[lt.InandOutputType.HEAT_CONSUMPTION],
    #             source_weight=999,
    #         )
    #     )
    #     return dynamic_connections

    def get_default_connections_from_more_advanced_heat_pump(
        self,
    ):
        """Get gas heater default connections."""

        from hisim.components.more_advanced_heat_pump_hplib import (  # pylint: disable=import-outside-toplevel
            MoreAdvancedHeatPumpHPLib,
        )

        dynamic_connections = []
        hp_class_name = MoreAdvancedHeatPumpHPLib.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=MoreAdvancedHeatPumpHPLib,
                source_class_name=hp_class_name,
                source_component_field_name=MoreAdvancedHeatPumpHPLib.ThermalOutputPowerTotal,
                source_load_type=lt.LoadTypes.HEATING,
                source_unit=lt.Units.WATT,
                source_tags=[lt.InandOutputType.HEAT_DELIVERED],
                source_weight=999,
            )
        )
        return dynamic_connections

    def get_default_connections_from_generic_heat_source(
        self,
    ):
        """Get generic heat source default connections."""

        from hisim.components.generic_heat_source import HeatSource  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        heat_source_class_name = HeatSource.get_classname()
        dynamic_connections.append(
            DynamicComponentConnection(
                source_component_class=HeatSource,
                source_class_name=heat_source_class_name,
                source_component_field_name=HeatSource.ThermalPowerDelivered,
                source_load_type=lt.LoadTypes.HEATING,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.InandOutputType.HEAT_DELIVERED,
                ],
                source_weight=999,
            )
        )
        return dynamic_connections

    def write_to_report(self):
        """Writes relevant information to report."""
        return self.grid_energy_balancer_config.get_string_dict()

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.self_copy()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.self_copy()

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks values."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the grid energy balancer."""

        if timestep == 0:
            self.production_inputs = self.get_dynamic_inputs(tags=[lt.InandOutputType.HEAT_DELIVERED])
            self.consumption_uncontrolled_inputs = self.get_dynamic_inputs(tags=[lt.InandOutputType.HEAT_CONSUMPTION])

        # get sum of production and consumption for all inputs for each iteration
        production_in_watt_hour = (
            sum([stsv.get_input_value(component_input=elem) for elem in self.production_inputs])
            * self.seconds_per_timestep
            / 3600
        )

        consumption_uncontrolled_in_watt_hour = (
            sum([stsv.get_input_value(component_input=elem) for elem in self.consumption_uncontrolled_inputs])
            * self.seconds_per_timestep
            / 3600
        )

        # Production of Heat positve sign
        # Consumption of Heat negative sign
        difference_between_production_and_consumption_in_watt_hour = (
            production_in_watt_hour - consumption_uncontrolled_in_watt_hour
        )

        # calculate cumulative production and consumption
        cumulative_production_in_watt_hour = self.state.cumulative_production_in_watt_hour + production_in_watt_hour
        cumulative_consumption_in_watt_hour = (
            self.state.cumulative_consumption_in_watt_hour + consumption_uncontrolled_in_watt_hour
        )

        # set outputs
        stsv.set_output_value(
            self.heat_available_channel,
            difference_between_production_and_consumption_in_watt_hour,
        )

        stsv.set_output_value(
            self.heat_consumption_channel,
            consumption_uncontrolled_in_watt_hour,
        )

        stsv.set_output_value(
            self.heat_production_channel,
            production_in_watt_hour,
        )

        stsv.set_output_value(
            self.cumulative_heat_consumption_channel,
            cumulative_consumption_in_watt_hour,
        )

        stsv.set_output_value(
            self.cumulative_heat_production_channel,
            cumulative_production_in_watt_hour,
        )

        self.state.cumulative_production_in_watt_hour = cumulative_production_in_watt_hour
        self.state.cumulative_consumption_in_watt_hour = cumulative_consumption_in_watt_hour

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of gas costs and revenues."""
        total_used_energy_in_kwh: float
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if output.field_name == self.HeatConsumption:
                    total_used_energy_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_used_energy_in_kwh = KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=total_used_energy_in_watt,
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    )

        emissions_and_cost_factors = EmissionFactorsAndCostsForFuelsConfig.get_values_for_year(
            self.my_simulation_parameters.year
        )
        co2_per_unit = emissions_and_cost_factors.contracting_heating_footprint_in_kg_per_kwh
        euro_per_unit = emissions_and_cost_factors.contracting_heating_costs_in_euro_per_kwh

        opex_cost_per_simulated_period_in_euro = total_used_energy_in_kwh * euro_per_unit
        co2_per_simulated_period_in_kg = total_used_energy_in_kwh * co2_per_unit
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=opex_cost_per_simulated_period_in_euro,
            opex_maintenance_cost_in_euro=0,
            co2_footprint_in_kg=co2_per_simulated_period_in_kg,
            consumption_in_kwh=total_used_energy_in_kwh,
            loadtype=lt.LoadTypes.GAS,
            kpi_tag=KpiTagEnumClass.HEATING_METER
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        total_used_energy_in_kwh: Optional[float] = None
        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name and output.load_type == lt.LoadTypes.HEATING:
                if output.field_name == self.HeatConsumption:
                    total_used_energy_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_used_energy_in_kwh = KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=total_used_energy_in_watt,
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    )

                    break

        total_heating_energy_consumption_in_building_in_kwh_entry = KpiEntry(
            name="Total heat consumption from grid",
            unit="kWh",
            value=total_used_energy_in_kwh,
            tag=KpiTagEnumClass.HEATING_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(total_heating_energy_consumption_in_building_in_kwh_entry)
        # try to get opex costs
        opex_costs = self.get_cost_opex(all_outputs=all_outputs, postprocessing_results=postprocessing_results)
        opex_costs_in_euro_entry = KpiEntry(
            name="Opex costs of heat consumption from grid",
            unit="Euro",
            value=opex_costs.opex_energy_cost_in_euro,
            tag=KpiTagEnumClass.HEATING_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(opex_costs_in_euro_entry)
        co2_footprint_in_kg_entry = KpiEntry(
            name="CO2 footprint of heat consumption from grid",
            unit="kg",
            value=opex_costs.co2_footprint_in_kg,
            tag=KpiTagEnumClass.HEATING_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(co2_footprint_in_kg_entry)

        return list_of_kpi_entries

    @staticmethod
    def get_cost_capex(config: HeatingMeterConfig, simulation_parameters: SimulationParameters) -> cp.CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = cp.CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class


@dataclass
class HeatingMeterState:
    """HeatingMeterState class."""

    cumulative_production_in_watt_hour: float
    cumulative_consumption_in_watt_hour: float

    def self_copy(
        self,
    ):
        """Copy the GasMeterState."""
        return HeatingMeterState(
            self.cumulative_production_in_watt_hour,
            self.cumulative_consumption_in_watt_hour,
        )



""" Modular Heat Pump Class together with Configuration and State. """

# clean

# Generic/Built-in
from dataclasses import dataclass
from typing import List, Optional

import pandas as pd
import numpy as np
from dataclasses_json import dataclass_json

import hisim.loadtypes as lt
from hisim import component as cp
from hisim.component import OpexCostDataClass, CapexCostDataClass

# Owned
from hisim import utils
from hisim.components import controller_l1_heatpump

from hisim.components.weather import Weather
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiHelperClass, KpiTagEnumClass

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class HeatPumpConfig(cp.ConfigBase):

    """Configuration of a HeatPump."""

    building_name: str
    #: name of the device
    name: str
    #: priority of the device in energy management system: the higher the number the lower the priority
    source_weight: int
    #: manufacturer to search heat pump in data base
    manufacturer: str
    #: device name to search heat pump in data base
    device_name: str
    #: maximal thermal power of heat pump in W
    power_th: float
    #: usage of the heatpump: either for heating or for water heating
    water_vs_heating: lt.InandOutputType
    #: category of the heat pump: either heat pump or heating rod
    device_category: lt.HeatingSystems
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    #: lifetime in years
    lifetime: float
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return ModularHeatPump.get_full_classname()

    @staticmethod
    def get_default_config_heating(
        building_name: str = "BUI1",
    ) -> "HeatPumpConfig":
        """Returns default configuration of a heat pump used for heating."""
        power_th: float = 6200  # W
        config = HeatPumpConfig(
            building_name=building_name,
            name="HeatingHeatPump",
            source_weight=1,
            manufacturer="Viessmann Werke GmbH & Co KG",
            device_name="Vitocal 300-A AWO-AC 301.B07",
            power_th=power_th,
            water_vs_heating=lt.InandOutputType.HEATING,
            device_category=lt.HeatingSystems.HEAT_PUMP,
            co2_footprint=power_th * 1e-3 * 165.84,  # value from emission_factros_and_costs_devices.csv
            cost=power_th * 1e-3 * 1513.74,  # value from emission_factros_and_costs_devices.csv
            lifetime=10,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )
        return config

    @staticmethod
    def get_default_config_waterheating(
        building_name: str = "BUI1",
    ) -> "HeatPumpConfig":
        """Returns default configuration of a heat pump used for water heating."""
        power_th: float = 3000  # W
        config = HeatPumpConfig(
            building_name=building_name,
            name="DHWHeatPump",
            source_weight=1,
            manufacturer="Viessmann Werke GmbH & Co KG",
            device_name="Vitocal 300-A AWO-AC 301.B07",
            power_th=power_th,
            water_vs_heating=lt.InandOutputType.WATER_HEATING,
            device_category=lt.HeatingSystems.HEAT_PUMP,
            co2_footprint=power_th * 1e-3 * 165.84,  # value from emission_factros_and_costs_devices.csv
            cost=power_th * 1e-3 * 1513.74,  # value from emission_factros_and_costs_devices.csv
            lifetime=10,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )
        return config

    @staticmethod
    def get_default_config_heating_electric(building_name: str = "BUI1",) -> "HeatPumpConfig":
        """Returns default configuartion of simple electrical heating system with a COP of one."""
        power_th: float = 6200  # W
        config = HeatPumpConfig(
            building_name=building_name,
            name="HeatingHeatingRod",
            source_weight=1,
            manufacturer="dummy",
            device_name="HeatingRod",
            power_th=power_th,
            water_vs_heating=lt.InandOutputType.HEATING,
            device_category=lt.HeatingSystems.ELECTRIC_HEATING,
            co2_footprint=power_th * 1e-3 * 1.21,  # value from emission_factros_and_costs_devices.csv
            cost=4635,  # value from emission_factros_and_costs_devices.csv
            lifetime=20,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )
        return config

    @staticmethod
    def get_default_config_waterheating_electric(
        building_name: str = "BUI1",
    ) -> "HeatPumpConfig":
        """Returns default configuration of electrical heating rod for boiler."""
        power_th: float = 3000  # W
        config = HeatPumpConfig(
            building_name=building_name,
            name="DHWHeatingRod",
            source_weight=1,
            manufacturer="dummy",
            device_name="HeatingRod",
            power_th=power_th,
            water_vs_heating=lt.InandOutputType.WATER_HEATING,
            device_category=lt.HeatingSystems.ELECTRIC_HEATING,
            co2_footprint=power_th * 1e-3 * 1.21,  # value from emission_factros_and_costs_devices.csv
            cost=4635,  # value from emission_factros_and_costs_devices.csv
            lifetime=20,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )
        return config

    @classmethod
    def get_scaled_waterheating_to_number_of_apartments(
        cls,
        number_of_apartments: float,
        default_power_in_watt: float = 3000,
        name: str = "DHWHeatPump",
        building_name: str = "BUI1",
    ) -> "HeatPumpConfig":
        """Gets a default heat pump with scaling according to number of apartments."""

        # scale with number of apartments
        power_th_in_watt: float = default_power_in_watt * number_of_apartments
        config = HeatPumpConfig(
            building_name=building_name,
            name=name,
            source_weight=1,
            manufacturer="Viessmann Werke GmbH & Co KG",
            device_name="Vitocal 300-A AWO-AC 301.B07",
            power_th=power_th_in_watt,
            water_vs_heating=lt.InandOutputType.WATER_HEATING,
            device_category=lt.HeatingSystems.HEAT_PUMP,
            co2_footprint=power_th_in_watt * 1e-3 * 165.84,  # value from emission_factros_and_costs_devices.csv
            cost=power_th_in_watt * 1e-3 * 1513.74,  # value from emission_factros_and_costs_devices.csv
            lifetime=10,  # value from emission_factros_and_costs_devices.csv
            maintenance_cost_as_percentage_of_investment=0.025,  # source:  VDI2067-1
        )
        return config


class ModularHeatPumpState:

    """Modular heat pump state saves the state of the heat pump."""

    def __init__(self, state: int = 0):
        """Initializes state."""
        self.state = state

    def clone(self) -> "ModularHeatPumpState":
        """Creates copy of state."""
        return ModularHeatPumpState(state=self.state)


class ModularHeatPump(cp.Component):

    """Heat pump implementation.

    The generic_heatpump_modular differs to generic_heatpump in the sense that the minimal runtime is not in the component,
    but in the related controller.
    This implementation does not consider cooling of building_names.

    Components to connect to:
    (1) Weather
    (2) Heat Pump Controller (controller_l1_heatpump)
    """

    # Inputs
    TemperatureOutside = "TemperatureOutside"
    HeatControllerTargetPercentage = "HeatControllerTargetPercentage"

    # Outputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    ElectricityOutput = "ElectricityOutput"
    PowerModifier = "PowerModifier"

    @utils.measure_execution_time
    def __init__(
        self,
        config: HeatPumpConfig,
        my_simulation_parameters: SimulationParameters,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ):
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.config = config
        self.build()
        self.state = ModularHeatPumpState()
        self.previous_state = ModularHeatPumpState()

        if my_simulation_parameters.surplus_control:
            postprocessing_flag = [
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
                self.config.water_vs_heating,
                self.config.device_category,
            ]
        else:
            postprocessing_flag = [
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED,
                self.config.water_vs_heating,
                self.config.device_category,
            ]

        # Inputs - Mandatories
        self.temperature_outside_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureOutside,
            lt.LoadTypes.ANY,
            lt.Units.CELSIUS,
            mandatory=True,
        )

        self.heat_controller_power_modifier_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.HeatControllerTargetPercentage,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            mandatory=True,
        )

        # Outputs
        self.thermal_power_delicered_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalPowerDelivered,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT,
            postprocessing_flag=[self.config.water_vs_heating],
            output_description="Thermal Power Delivered",
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=postprocessing_flag,
            output_description="Electricity Output",
        )

        self.power_modifier_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.PowerModifier,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.ANY,
            postprocessing_flag=[],
            output_description="Power Modifier",
        )

        self.add_default_connections(self.get_default_connections_from_weather())
        self.add_default_connections(self.get_default_connections_from_controller_l1_heatpump())

    def get_default_connections_from_weather(self):
        """Sets default connections of Weather."""

        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                ModularHeatPump.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        return connections

    def get_default_connections_from_controller_l1_heatpump(self):
        """Sets default connections of heat pump controller."""

        connections = []
        controller_classname = controller_l1_heatpump.L1HeatPumpController.get_classname()
        connections.append(
            cp.ComponentConnection(
                ModularHeatPump.HeatControllerTargetPercentage,
                controller_classname,
                controller_l1_heatpump.L1HeatPumpController.HeatControllerTargetPercentage,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def build(self):
        """Initialization function of Modular Heat Pump."""
        # Retrieves heat pump from database - BEGIN
        heat_pumps_database = utils.load_smart_appliance("Heat Pump")

        heat_pump_found = False
        heat_pump = None
        for heat_pump in heat_pumps_database:
            if heat_pump["Manufacturer"] == self.config.manufacturer and heat_pump["Name"] == self.config.device_name:
                heat_pump_found = True
                break

        if not heat_pump_found or heat_pump is None:
            raise Exception("Heat pump model not registered in the database")

        # Interpolates COP data from the database
        self.cop_ref = []
        self.t_out_ref = []
        for heat_pump_cops in heat_pump["COP"]:
            self.t_out_ref.append(float([*heat_pump_cops][0][1:].split("/")[0]))
            self.cop_ref.append(float([*heat_pump_cops.values()][0]))
        self.cop_coef = np.polyfit(self.t_out_ref, self.cop_ref, 1)

        # Writes info to report
        self.write_to_report()

    def cal_cop(self, t_out: float) -> float:
        """Returns coefficient of performance of selected heat pump."""
        val: float = self.cop_coef[0] * t_out + self.cop_coef[1]
        return val

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def write_to_report(self) -> List[str]:
        """Writes relevant data to report."""
        return self.config.get_string_dict()

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Iteration of heat pump simulation."""

        # Inputs
        target_percentage = stsv.get_input_value(self.heat_controller_power_modifier_channel)

        temperature_outside: float = stsv.get_input_value(self.temperature_outside_channel)
        cop = self.cal_cop(temperature_outside)
        electric_power = self.config.power_th / cop

        # calculate modulation
        if target_percentage > 0:
            power_modifier = target_percentage
        elif target_percentage == 0:
            power_modifier = 0
        else:
            raise ValueError("`target_modifiert` needs to be a positive number.")

        power_modifier = min(1, power_modifier)

        stsv.set_output_value(self.thermal_power_delicered_channel, self.config.power_th * power_modifier)
        stsv.set_output_value(self.power_modifier_channel, power_modifier)

        stsv.set_output_value(self.electricity_output_channel, electric_power * power_modifier)

    @staticmethod
    def get_cost_capex(config: HeatPumpConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        seconds_per_year = 365 * 24 * 60 * 60
        capex_per_simulated_period = (config.cost / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        device_co2_footprint_per_simulated_period = (config.co2_footprint / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )

        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost,
            device_co2_footprint_in_kg=config.co2_footprint,
            lifetime_in_years=config.lifetime,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.HEATPUMP_DOMESTIC_HOT_WATER
        )
        return capex_cost_data_class

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of maintenance costs snd write total energy consumption to component-config.

        No electricity costs for components except for Electricity Meter,
        because part of electricity consumption is feed by PV
        """
        #: consumption of the heatpump in kWh
        consumption_in_kwh: float

        for index, output in enumerate(all_outputs):
            if (
                    output.component_name == self.component_name
                    and output.load_type == lt.LoadTypes.ELECTRICITY and output.field_name == self.ElectricityOutput
            ):  # Todo: check component name from system_setups: find another way of using only heatpump-outputs
                consumption_in_kwh = round(
                    sum(postprocessing_results.iloc[:, index])
                    * self.my_simulation_parameters.seconds_per_timestep
                    / 3.6e6,
                    1,
                )

        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=0,
            consumption_in_kwh=consumption_in_kwh,
            loadtype=lt.LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.HEATPUMP_DOMESTIC_HOT_WATER
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        dhw_heat_pump_total_electricity_consumption_in_kilowatt_hour: Optional[float] = None
        dhw_heat_pump_heating_energy_output_in_kilowatt_hour: Optional[float] = None
        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if output.field_name == self.ElectricityOutput:
                    dhw_heat_pump_total_electricity_consumption_in_watt_series = postprocessing_results.iloc[:, index]
                    dhw_heat_pump_total_electricity_consumption_in_kilowatt_hour = (
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=dhw_heat_pump_total_electricity_consumption_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                elif output.field_name == self.ThermalPowerDelivered:
                    dhw_heat_pump_heating_power_output_in_watt_series = postprocessing_results.iloc[:, index]
                    dhw_heat_pump_heating_energy_output_in_kilowatt_hour = (
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=dhw_heat_pump_heating_power_output_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )

        dhw_heatpump_total_electricity_consumption_entry = KpiEntry(
            name="DHW heat pump total electricity consumption",
            unit="kWh",
            value=dhw_heat_pump_total_electricity_consumption_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_DOMESTIC_HOT_WATER,
            description=self.component_name
        )
        list_of_kpi_entries.append(dhw_heatpump_total_electricity_consumption_entry)

        dhw_heatpump_heating_energy_output_entry = KpiEntry(
            name="Heating output energy of DHW heat pump",
            unit="kWh",
            value=dhw_heat_pump_heating_energy_output_in_kilowatt_hour,
            tag=KpiTagEnumClass.HEATPUMP_DOMESTIC_HOT_WATER,
            description=self.component_name
        )
        list_of_kpi_entries.append(dhw_heatpump_heating_energy_output_entry)

        return list_of_kpi_entries



""" Generates a Json file for the json executor. """
# clean
from typing import Any, Optional, List, Type, Dict, Union
import json
from dataclasses import dataclass, field
from dataclass_wizard import JSONWizard

from hisim import log
from hisim.component import ConfigBase
from hisim.simulationparameters import SimulationParameters


@dataclass
class ConnectionEntry(JSONWizard):

    """Represents a connection between two components."""

    src_component_name: str
    src_output_name: str
    dst_input_name: str


@dataclass
class ComponentEntry(JSONWizard):

    """Entry for a single component."""

    component_full_classname: str
    component_name: str
    configuration: Dict[Any, Any]
    config_full_classname: str
    default_connections: List[str]
    manual_connections: List[ConnectionEntry]


@dataclass
class ConfigFile(JSONWizard):

    """Contains all the data that goes into the json config file."""

    system_name: str
    component_entries: List[ComponentEntry] = field(default_factory=list)
    my_simulation_parameters: Optional[SimulationParameters] = None
    my_module_config: Optional[Dict[str, Any]] = None
    scenario_data_information: Optional[Dict[str, Any]] = None


class JsonConfigurationGenerator:

    """Class to generate Json config files that can be used to start calculations without Python."""

    def __init__(self, name: str) -> None:
        """Initializes the configuration generator."""
        self.config_file: ConfigFile = ConfigFile(system_name=name)

    def set_simulation_parameters(
        self, my_simulation_parameters: SimulationParameters
    ) -> None:
        """Sets the simulation parameters."""
        self.config_file.my_simulation_parameters = my_simulation_parameters

    def set_module_config(self, my_module_config: Union[str, Dict]) -> None:
        """Sets the module config which is was used to configure the example module."""
        if isinstance(my_module_config, str):
            with open(my_module_config.rstrip("\r"), encoding="unicode_escape") as openfile:  # type: ignore
                config_dict = json.load(openfile)
        elif isinstance(my_module_config, Dict):
            config_dict = my_module_config
        else:
            raise TypeError(f"The passed argument has type {type(my_module_config)} but a str or Dict is expected.")
        self.config_file.my_module_config = config_dict

    def set_scenario_data_information_dict(
        self, scenario_data_information_dict: Dict[str, Any]
    ) -> None:
        """Sets some scenario information concerning the result data."""
        self.config_file.scenario_data_information = scenario_data_information_dict

    def add_component(self, config: Type[ConfigBase]) -> ComponentEntry:
        """Adds a component and returns a component entry."""
        config_json_str = config.to_dict()
        component_entry = ComponentEntry(
            component_full_classname=config.get_main_classname(),
            component_name=config.name,
            configuration=config_json_str,
            config_full_classname=config.get_config_classname(),
            default_connections=[],
            manual_connections=[],
        )
        self.config_file.component_entries.append(component_entry)
        log.information("Added component " + config.name)
        return component_entry

    def add_default_connection(
        self, from_entry: ComponentEntry, to_entry: ComponentEntry
    ) -> None:
        """Adds a default connection to another component."""
        to_entry.default_connections.append(from_entry.component_name)

    def add_manual_connection(
        self,
        from_entry: ComponentEntry,
        output_name: str,
        to_entry: ComponentEntry,
        input_name: str,
    ) -> None:
        """Adds a manual connection entry for connecting a single input."""
        my_connection = ConnectionEntry(
            from_entry.component_name, output_name, input_name
        )
        to_entry.manual_connections.append(my_connection)

    def save_to_json(self, filename: str) -> None:
        """Saves a configuration to a json file."""
        with open(filename, "w", encoding="utf-8") as filestream:
            mystr = self.config_file.to_json(indent=4)
            filestream.write(mystr)



"""KPI config module."""

from dataclasses import dataclass
import enum
from dataclasses_json import dataclass_json


@enum.unique
class KPIForRatingInOptimization(str, enum.Enum):

    """Choose KPI that will be optimized with building sizer."""

    TOTAL_COSTS = "Total Costs [€]"
    ENERGY_GRID_COSTS = "Energy Grid Costs [€]"
    MAINTENANCE_COSTS = "Maintenance Costs [€]"
    INVESTMENT_COSTS = "Investment Costs [€]"
    TOTAL_CO2_EMISSION = "Total CO2 Emission [kg]"
    SELFSUFFICIENCY = "Self-sufficiency rate [%]"


@dataclass_json
@dataclass
class KPIConfig:

    """KPI config class."""

    #: ratio between the load covered onsite and the total load, given in %
    self_sufficiency_rate_in_percent: float
    #: annual cost for investment and operation in the considered technology, given in euros
    total_costs_in_euro: float
    #: annual cost for energy from grid (electricty, gas, heat) given in euros
    energy_grid_costs_in_euro: float
    #: annual cost for maintenance given in euros
    maintenance_costs_in_euro: float
    #: annual cost for investment given in euros
    investment_costs_in_euro: float
    #: annual C02 emmissions due to the construction and operation of the considered technology, given in kg
    total_co2_emissions_in_kg: float

    def get_kpi_for_rating(self, chosen_kpi: KPIForRatingInOptimization) -> float:
        """Weights all kpis to get one value evaluating the performance of one building configuration.

        Also referred to as "rating" or "fitness" in the evolutionary algorithm of the building sizer.
        """

        if chosen_kpi == KPIForRatingInOptimization.SELFSUFFICIENCY:
            return self.self_sufficiency_rate_in_percent
        if chosen_kpi == KPIForRatingInOptimization.TOTAL_COSTS:
            return self.total_costs_in_euro
        if chosen_kpi == KPIForRatingInOptimization.ENERGY_GRID_COSTS:
            return self.energy_grid_costs_in_euro
        if chosen_kpi == KPIForRatingInOptimization.MAINTENANCE_COSTS:
            return self.maintenance_costs_in_euro
        if chosen_kpi == KPIForRatingInOptimization.INVESTMENT_COSTS:
            return self.investment_costs_in_euro
        if chosen_kpi == KPIForRatingInOptimization.TOTAL_CO2_EMISSION:
            return self.total_co2_emissions_in_kg
        raise ValueError(f"Chosen KPI {self.chosen_kpi} not recognized.")



"""System setup configuration module."""

# clean

from typing import Any
import json
import os
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
from typing_extensions import Self

from hisim import log, utils
from hisim.components import building


@dataclass
class SystemSetupConfigBase(JSONWizard):

    """Base class for system setup."""

    @classmethod
    def load_from_json(cls, module_config_path: str) -> Self:
        """Populate config from JSON."""

        if os.path.exists(module_config_path):
            with open(module_config_path, "r", encoding="utf8") as file:
                module_config_dict = json.loads(file.read())
        elif os.path.exists(module_config_path.rstrip("\r")):
            with open(module_config_path.rstrip("\r"), "r", encoding="utf8") as file:
                module_config_dict = json.loads(file.read())
        else:
            raise FileExistsError(f"The module config file {module_config_path} could not be found.")
        log.information(f"Read module config from {module_config_path}.")

        # Read building config overwrites. It is used to scale the system setup.
        building_config_dict = module_config_dict.pop("building_config", {})
        if building_config_dict:
            log.information("Using `building_config` for scaling.")
            building_config = building.BuildingConfig.from_dict(building_config_dict)
        else:
            building_config = None

        # Load option overwrites.
        options_dict = module_config_dict.pop("options", {})
        options = cls.get_default_options()
        if options_dict:
            log.information("Using `options`.")
            utils.set_attributes_of_dataclass_from_dict(options, options_dict)

        # Load (scaled) default values for system setup configuration.
        if options_dict and not building_config:
            raise ValueError("Options for default setup not yet implemented.")
        if building_config:
            my_config = cls.get_scaled_default(building_config=building_config, options=options)
        else:
            my_config = cls.get_default()

        # Read setup config overwrites
        setup_config_dict = module_config_dict.pop("system_setup_config", {})
        if setup_config_dict:
            log.information("Using `system_setup_config` to overwrite defaults.")
            utils.set_attributes_of_dataclass_from_dict(my_config, setup_config_dict)
        else:
            log.information("Did not find `system_setup_config` in JSON. Using defaults.")

        return my_config

    @classmethod
    def get_default_options(cls) -> Any:
        """Get default options."""
        raise NotImplementedError

    @classmethod
    def get_default(cls) -> Self:
        """Get default."""
        raise NotImplementedError

    @classmethod
    def get_scaled_default(cls, building_config: building.BuildingConfig, options: Any) -> Self:
        """Get scaled default."""
        raise NotImplementedError



"""Controller of the generic_electrolyzer.

The controller looks at the available surplus electricity and passes the signal to the electrolyzer accordingly.
In addition the controller takes care of minimum operation and indle times and the available capacity in the hydrogen storage.
"""

# clean
import importlib
from dataclasses import dataclass
from typing import List
from dataclasses_json import dataclass_json

from hisim import utils
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters


@dataclass_json
@dataclass
class L1ElectrolyzerControllerConfig(cp.ConfigBase):
    """Electrolyzer Controller Config."""

    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    # minimal operation time of heat source
    min_operation_time_in_seconds: int
    # minimal resting time of heat source
    min_idle_time_in_seconds: int
    #: minimal electrical power of the electrolyzer
    p_min_electrolyzer: float
    #: maximal allowed content of hydrogen storage for turning the electrolyzer on
    h2_soc_threshold: float

    @staticmethod
    def get_default_config(
        building_name: str = "BUI1",
    ) -> "L1ElectrolyzerControllerConfig":
        """Returns the default configuration of an electrolyzer controller."""
        config = L1ElectrolyzerControllerConfig(
            building_name=building_name,
            name="L1 Electrolyzer Controller",
            source_weight=1,
            min_operation_time_in_seconds=14400,
            min_idle_time_in_seconds=7200,
            p_min_electrolyzer=1200,
            h2_soc_threshold=96,
        )
        return config


class L1ElectrolyzerControllerState:
    """Data class for saving the state of the electrolyzer controller."""

    def __init__(
        self,
        state: int = 0,
        activation_time_step: int = 0,
        deactivation_time_step: int = 0,
    ):
        """Initialize the class."""
        self.state: int = state
        self.activation_time_step: int = activation_time_step
        self.deactivation_time_step: int = deactivation_time_step

    def clone(self) -> "L1ElectrolyzerControllerState":
        """Clone function."""
        return L1ElectrolyzerControllerState(
            state=self.state,
            activation_time_step=self.activation_time_step,
            deactivation_time_step=self.deactivation_time_step,
        )

    def activate(self, timestep: int) -> None:
        """Activates the heat pump and remembers the time step."""
        if self.state == 0:
            self.activation_time_step = timestep
        self.state = 1

    def deactivate(self, timestep: int) -> None:
        """Deactivates the heat pump and remembers the time step."""
        if self.state == 1:
            self.deactivation_time_step = timestep
        self.state = 0


class L1GenericElectrolyzerController(cp.Component):
    """Controller of the Electrolyzer.

    It takes the available surplus electricity of the energy management system and passes it to the electrolyzer.
    If either the available surplus is too low, or the availabel capacity of the hydrogen storage is below the indicated threshold,
    the electricity surplus signal is modified accordingly.

    Components to connect to:
    (1) energy management system (controller_l2_energy_management_system)
    (2) hydrogen storage (generic_h2storage)
    """

    # Inputs
    ElectricityTarget = "ElectricityTarget"
    HydrogenSOC = "HydrogenSOC"

    # Outputs
    AvailableElectricity = "AvailableElectricity"

    # Similar components to connect to:
    # 1. building_name
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: L1ElectrolyzerControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.minimum_runtime_in_timesteps = int(
            config.min_operation_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )
        self.minimum_resting_time_in_timesteps = int(
            config.min_idle_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )

        self.state = L1ElectrolyzerControllerState()
        self.previous_state = L1ElectrolyzerControllerState()
        self.processed_state = L1ElectrolyzerControllerState()

        # add inputs
        self.electricity_target_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityTarget,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            mandatory=True,
        )
        self.hydrogen_soc_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.HydrogenSOC,
            lt.LoadTypes.HYDROGEN,
            lt.Units.PERCENT,
            mandatory=True,
        )
        # add outputs
        self.available_electicity_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.AvailableElectricity,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description="Available Electricity for Electrolyzer from Electrolyzer Controller.",
        )

        self.add_default_connections(self.get_default_connections_from_h2_storage())

    def get_default_connections_from_h2_storage(self):
        """Sets default connections for the hydrogen storage in the electrolyzer controller."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_hydrogen_storage"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "GenericHydrogenStorage")
        connections = []
        h2_storage_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1GenericElectrolyzerController.HydrogenSOC,
                h2_storage_classname,
                component_class.HydrogenSOC,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Save the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restore the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""

        if force_convergence:
            electricity_target = stsv.get_input_value(self.electricity_target_channel)
            self.state = self.processed_state
        else:
            electricity_target = stsv.get_input_value(self.electricity_target_channel)
            h2_soc = stsv.get_input_value(self.hydrogen_soc_channel)
            self.calculate_state(timestep, electricity_target, h2_soc)
            self.processed_state = self.state.clone()

        # minimum power of electrolyzer fulfilled when running
        if self.state.state == 1 and electricity_target < self.config.p_min_electrolyzer:
            electricity_target = self.config.p_min_electrolyzer
        stsv.set_output_value(
            self.available_electicity_output_channel,
            self.state.state * electricity_target,
        )

    def calculate_state(self, timestep: int, electricity_target: float, h2_soc: float) -> None:
        """Calculate the state."""
        # return device on if minimum operation time is not fulfilled and device was on in previous state
        if self.state.state == 1 and self.state.activation_time_step + self.minimum_runtime_in_timesteps >= timestep:
            # mandatory on, minimum runtime not reached
            return
        if (
            self.state.state == 0
            and self.state.deactivation_time_step + self.minimum_resting_time_in_timesteps >= timestep
        ):
            # mandatory off, minimum resting time not reached
            return

        # available electricity too low or hydrogen storage too full
        if (electricity_target < self.config.p_min_electrolyzer) or (h2_soc > self.config.h2_soc_threshold):
            self.state.deactivate(timestep)
            return

        # turns on if electricity is high enough and there is still space in storage
        self.state.activate(timestep)

    def write_to_report(self) -> List[str]:
        """Writes the information of the current component to the report."""
        return self.config.get_string_dict()



"""Simple implementation of combined heat and power plant (CHP).

This can be either a natural gas driven turbine producing both electricity and heat,
or also a fuel cell. In this implementation the CHP does not modulate: it is either
on or off. When it runs, it outputs a constant thermal and electrical power signal
and needs a constant input of hydrogen or natural gas.
"""

# clean

from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.components import controller_l1_chp
from hisim.simulationparameters import SimulationParameters

__authors__ = "Frank Burkrad, Maximilian Hillen,"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = "development"


@dataclass_json
@dataclass
class CHPConfig(cp.ConfigBase):
    """Defininition of configuration of combined heat and power plant (CHP)."""

    building_name: str
    #: name of the CHP
    name: str
    #: priority of the component in hierachy: the higher the number the lower the priority
    source_weight: int
    #: type of CHP (fuel cell or gas driven)
    use: lt.LoadTypes
    #: electrical power of the CHP, when activated
    p_el: float
    #: thermal power of the CHP in Watt, when activated
    p_th: float
    #: demanded power of fuel input in Watt, when activated
    p_fuel: float

    @staticmethod
    def get_default_config_chp(
        thermal_power: float,
        building_name: str = "BUI1",
    ) -> "CHPConfig":
        """Get default config chp."""
        config = CHPConfig(
            building_name=building_name,
            name="CHP",
            source_weight=1,
            use=lt.LoadTypes.GAS,
            p_el=(0.33 / 0.5) * thermal_power,
            p_th=thermal_power,
            p_fuel=(1 / 0.5) * thermal_power,
        )
        return config

    @staticmethod
    def get_default_config_fuelcell(
        thermal_power: float,
        building_name: str = "BUI1",
    ) -> "CHPConfig":
        """Get default config fuel cell."""
        config = CHPConfig(
            building_name=building_name,
            name="CHP",
            source_weight=1,
            use=lt.LoadTypes.HYDROGEN,
            p_el=(0.48 / 0.43) * thermal_power,
            p_th=thermal_power,
            p_fuel=(1 / 0.43) * thermal_power,
        )
        return config


class GenericCHPState:
    """Generic chp state class saves the state of the CHP."""

    def __init__(self, state: int) -> None:
        """Initialize the class."""
        self.state = state

    def clone(self) -> "GenericCHPState":
        """Clones the state."""
        return GenericCHPState(state=self.state)


class SimpleCHP(cp.Component):
    """Simulates CHP operation with constant electical and thermal power as well as constant fuel consumption.

    Components to connect to:
    (1) CHP or fuel cell controller (controller_l1_chp)
    """

    # Inputs
    CHPControllerOnOffSignal = "CHPControllerOnOffSignal"
    CHPControllerHeatingModeSignal = "CHPControllerHeatingModeSignal"

    # Outputs
    ThermalPowerOutputBuilding = "ThermalPowerOutputBuilding"
    ThermalPowerOutputBoiler = "ThermalPowerOutputBoiler"
    ElectricityOutput = "ElectricityOutput"
    FuelDelivered = "FuelDelivered"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: CHPConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initializes the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.config = config
        if self.config.use == lt.LoadTypes.HYDROGEN:
            self.p_fuel = config.p_fuel / (3.6e3 * 3.939e4)  # converted to kg / s
        else:
            self.p_fuel = config.p_fuel * my_simulation_parameters.seconds_per_timestep / 3.6e3  # converted to Wh

        self.state = GenericCHPState(state=0)
        self.previous_state = self.state.clone()

        # Inputs
        self.chp_onoff_signal_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.CHPControllerOnOffSignal,
            lt.LoadTypes.ON_OFF,
            lt.Units.BINARY,
            mandatory=True,
        )

        self.chp_heatingmode_signal_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.CHPControllerHeatingModeSignal,
            lt.LoadTypes.ANY,
            lt.Units.BINARY,
            mandatory=True,
        )

        # Component outputs
        self.thermal_power_output_building_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalPowerOutputBuilding,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT,
            postprocessing_flag=[lt.InandOutputType.THERMAL_PRODUCTION],
            output_description="Thermal Power output from CHP to building or buffer in Watt.",
        )
        self.thermal_power_output_dhw_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ThermalPowerOutputBoiler,
            load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT,
            postprocessing_flag=[lt.InandOutputType.THERMAL_PRODUCTION],
            output_description="Thermal Power output from CHP to drain hot water storage in Watt.",
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_PRODUCTION,
                lt.ComponentType.FUEL_CELL,
            ],
            output_description="Electrical Power output of CHP in Watt.",
        )
        if self.config.use == lt.LoadTypes.HYDROGEN:
            self.fuel_consumption_channel: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.FuelDelivered,
                load_type=lt.LoadTypes.HYDROGEN,
                unit=lt.Units.KG_PER_SEC,
                postprocessing_flag=[lt.LoadTypes.HYDROGEN],
                output_description="Hydrogen consumption of CHP in kg / s.",
            )
        elif self.config.use == lt.LoadTypes.GAS:
            self.fuel_consumption_channel = self.add_output(
                object_name=self.component_name,
                field_name=self.FuelDelivered,
                load_type=lt.LoadTypes.GAS,
                unit=lt.Units.WATT_HOUR,
                postprocessing_flag=[
                    lt.InandOutputType.FUEL_CONSUMPTION,
                    lt.LoadTypes.GAS,
                ],
                output_description="Gas consumption of CHP in Wh.",
            )
        self.add_default_connections(self.get_default_connections_from_chp_controller())

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        # Inputs
        self.state.state = int(stsv.get_input_value(self.chp_onoff_signal_channel))
        mode = stsv.get_input_value(self.chp_heatingmode_signal_channel)

        # Outputs
        if mode == 0:
            stsv.set_output_value(
                self.thermal_power_output_dhw_channel,
                self.state.state * self.config.p_th,
            )
            stsv.set_output_value(self.thermal_power_output_building_channel, 0)
        elif mode == 1:
            stsv.set_output_value(self.thermal_power_output_dhw_channel, 0)
            stsv.set_output_value(
                self.thermal_power_output_building_channel,
                self.state.state * self.config.p_th,
            )

        stsv.set_output_value(self.electricity_output_channel, self.state.state * self.config.p_el)
        stsv.set_output_value(self.fuel_consumption_channel, self.state.state * self.p_fuel)

    def get_default_connections_from_chp_controller(
        self,
    ) -> List[cp.ComponentConnection]:
        """Sets default connections of the controller in the Fuel Cell / CHP."""

        connections: List[cp.ComponentConnection] = []
        controller_classname = controller_l1_chp.L1CHPController.get_classname()
        connections.append(
            cp.ComponentConnection(
                SimpleCHP.CHPControllerOnOffSignal,
                controller_classname,
                controller_l1_chp.L1CHPController.CHPControllerOnOffSignal,
            )
        )
        connections.append(
            cp.ComponentConnection(
                SimpleCHP.CHPControllerHeatingModeSignal,
                controller_classname,
                controller_l1_chp.L1CHPController.CHPControllerHeatingModeSignal,
            )
        )
        return connections

    def write_to_report(self):
        """Writes the information of the current component to the report."""
        return self.config.get_string_dict()



"""Module for generating random numbers."""

# clean

# Generic/Built-in
import random
from typing import List
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Owned
from hisim.component import Component, SingleTimeStepValues, ConfigBase, DisplayConfig
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters


@dataclass_json
@dataclass
class RandomNumbersConfig(ConfigBase):
    """Configuration of the Random Numbers."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return RandomNumbers.get_full_classname()

    building_name: str
    name: str
    timesteps: int
    minimum: float
    maximum: float

    @classmethod
    def get_default_config(cls):
        """Gets a default config."""
        return RandomNumbersConfig(
            building_name="BUI1",
            name="RandomNumbers",
            timesteps=100,
            minimum=1,
            maximum=20,
        )


class RandomNumbers(Component):
    """Random number class."""

    RandomOutput: str = "Random Numbers"

    def __init__(
        self,
        config: RandomNumbersConfig,
        my_simulation_parameters: SimulationParameters,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.values: List[float] = []
        self.minimum = config.minimum
        self.maximum = config.maximum
        number_range = config.maximum - config.minimum
        for _ in range(config.timesteps):
            number = config.minimum + random.random() * number_range
            self.values.append(number)
        self.output1 = self.add_output(
            self.component_name,
            RandomNumbers.RandomOutput,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Random Number Output",
        )

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""

        val1: float = self.values[timestep]
        stsv.set_output_value(self.output1, float(val1))

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def write_to_report(self) -> List[str]:
        """Write to report."""
        lines = []
        lines.append(f"Random number Generator: {self.component_name}")
        lines.append(f"Minimum number: {self.minimum}")
        lines.append(f"Maximum number: {self.maximum}")
        return lines



""" Handles all the weather data processing. """

# clean
import csv
import datetime
import math
import os
from dataclasses import dataclass
from enum import Enum
from typing import Any, List

import numpy as np
import pandas as pd
import pvlib

from hisim import loadtypes as lt
from hisim import log, utils
from hisim.component import Component, ComponentOutput, ConfigBase, SingleTimeStepValues, DisplayConfig, OpexCostDataClass, CapexCostDataClass
from hisim.simulationparameters import SimulationParameters
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry

__authors__ = "Vitor Hugo Bellotto Zago, Noah Pflugradt"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Noah Pflugradt"

""" The functions cited in this module are at some degree based on the tsib project:

[tsib-kotzur]: Kotzur, Leander, Detlef Stolten, and Hermann-Josef Wagner. Future grid load of the residential building sector.
No. RWTH-2018-231872. Lehrstuhl für Brennstoffzellen (FZ Jülich), 2019.
ID: http://hdl.handle.net/2128/21115
    http://nbn-resolving.org/resolver?verb=redirect&identifier=urn:nbn:de:0001-2019020614

The implementation of the tsib project can be found under the following repository:
https://github.com/FZJ-IEK3-VSA/tsib
"""


class WeatherDataSourceEnum(Enum):
    """Describes where the weather data is from. Used to choose the correct reading function."""

    DWD_TRY = 1
    NSRDB = 2
    NSRDB_15MIN = 3
    DWD_10MIN = 4
    ERA5 = 5


class LocationEnum(Enum):
    """contains all the locations and their corresponding directories."""

    AACHEN = (
        "Aachen",
        "test-reference-years_1995-2012_1-location",
        "data_processed",
        "aachen_center",
        WeatherDataSourceEnum.DWD_TRY,
    )
    BREMERHAVEN = (
        "01_Bremerhaven",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_01",
        WeatherDataSourceEnum.DWD_TRY,
    )
    ROSTOCK = (
        "02_Rostock",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_02",
        WeatherDataSourceEnum.DWD_TRY,
    )
    HAMBURG = (
        "03Hamburg",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_03",
        WeatherDataSourceEnum.DWD_TRY,
    )
    POTSDAM = (
        "04Potsdam",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_04",
        WeatherDataSourceEnum.DWD_TRY,
    )
    ESSEN = (
        "05Essen",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_05",
        WeatherDataSourceEnum.DWD_TRY,
    )
    BAD_MARIENBURG = (
        "06Bad Marienburg",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_06",
        WeatherDataSourceEnum.DWD_TRY,
    )
    KASSEL = (
        "07Kassel",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_07",
        WeatherDataSourceEnum.DWD_TRY,
    )
    BRAUNLAGE = (
        "08Braunlage",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_08",
        WeatherDataSourceEnum.DWD_TRY,
    )
    CHEMNITZ = (
        "09Chemnitz",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_09",
        WeatherDataSourceEnum.DWD_TRY,
    )
    HOF = (
        "10Hof",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_10",
        WeatherDataSourceEnum.DWD_TRY,
    )
    FICHTELBERG = (
        "11Fichtelberg",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_11",
        WeatherDataSourceEnum.DWD_TRY,
    )
    MANNHEIM = (
        "12Mannheim",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_12",
        WeatherDataSourceEnum.DWD_TRY,
    )
    MUEHLDORF = (
        "13Muehldorf",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_13",
        WeatherDataSourceEnum.DWD_TRY,
    )
    STOETTEN = (
        "14Stoetten",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_14",
        WeatherDataSourceEnum.DWD_TRY,
    )
    GARMISCH_PARTENKIRCHEN = (
        "15Garmisch Partenkirchen",
        "test-reference-years_2015-2045_15-locations",
        "data_processed",
        "weather_region_15",
        WeatherDataSourceEnum.DWD_TRY,
    )
    MADRID = (
        "Madrid",
        "NSRDB",
        "Madrid",
        "Madrid",
        WeatherDataSourceEnum.NSRDB,
    )
    SEVILLE = (
        "Seville",
        "NSRDB",
        "Seville",
        "Seville",
        WeatherDataSourceEnum.NSRDB,
    )
    ATHENS = (
        "Athens",
        "NSRDB",
        "Athens",
        "Athens",
        WeatherDataSourceEnum.NSRDB,
    )
    BELGRADE = (
        "Belgrade",
        "NSRDB",
        "Belgrade",
        "Belgrade",
        WeatherDataSourceEnum.NSRDB,
    )
    CYPRUS = (
        "Cyprus",
        "NSRDB",
        "Cyprus",
        "Cyprus",
        WeatherDataSourceEnum.NSRDB,
    )
    LJUBLIANA = (
        "Ljubljana",
        "NSRDB",
        "Ljubljana",
        "Ljubljana",
        WeatherDataSourceEnum.NSRDB,
    )
    MILAN = (
        "Milan",
        "NSRDB",
        "Milan",
        "Milan",
        WeatherDataSourceEnum.NSRDB,
    )
    SARAJEVO = (
        "Sarajevo",
        "NSRDB",
        "Sarajevo",
        "Sarajevo",
        WeatherDataSourceEnum.NSRDB,
    )
    VRANJE = (
        "Vranje",
        "NSRDB",
        "Vranje",
        "Vranje",
        WeatherDataSourceEnum.NSRDB,
    )
    FR = (
        "Paris",
        "NSRDB_15min",
        "Paris",
        "403286_48.85_2.34_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    DE = (
        "Potsdam",
        "NSRDB_15min",
        "Potsdam",
        "742114_52.41_13.06_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    PL = (
        "Warsaw",
        "NSRDB_15min",
        "Warsaw",
        "1138443_52.25_21.02_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    SE = (
        "Stockholm",
        "NSRDB_15min",
        "Stockholm",
        "984998_59.33_18.06_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    NO = (
        "Oslo",
        "NSRDB_15min",
        "Oslo",
        "653025_59.93_10.74_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    RS = (
        "Belgrad",
        "NSRDB_15min",
        "Belgrad",
        "1108363_44.77_20.46_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    IT = (
        "Rome",
        "NSRDB_15min",
        "Rome",
        "718838_41.89_12.50_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    GB = (
        "London",
        "NSRDB_15min",
        "London",
        "337089_51.49_-0.10_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    CY = (
        "Nicosia",
        "NSRDB_15min",
        "Nicosia",
        "1809004_35.17_33.38_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    GR = (
        "Athens",
        "NSRDB_15min",
        "Athens",
        "1291832_37.97_23.74_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    IE = (
        "Dublin",
        "NSRDB_15min",
        "Dublin",
        "165308_53.37_-6.26_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    SI = (
        "Ljubljana",
        "NSRDB_15min",
        "Ljubljana",
        "808557_46.05_14.50_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    CZ = (
        "Prague",
        "NSRDB_15min",
        "Prague",
        "804583_50.09_14.42_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    AT = (
        "Viena",
        "NSRDB_15min",
        "Viena",
        "902141_48.21_16.38_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    HU = (
        "Budapest",
        "NSRDB_15min",
        "Budapest",
        "1035927_47.49_19.06_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    BE = (
        "Uccle",
        "NSRDB_15min",
        "Uccle",
        "454992_50.81_4.34_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    ES = (
        "Malaga",
        "NSRDB_15min",
        "Malaga",
        "213028_36.73_-4.42_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    DK = (
        "Copenhagen",
        "NSRDB_15min",
        "Copenhagen",
        "721796_55.69_12.58_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    NL = (
        "Amsterdam",
        "NSRDB_15min",
        "Amsterdam",
        "469536_52.37_4.90_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )
    BG = (
        "Sofia",
        "NSRDB_15min",
        "Sofia",
        "1267064_42.69_23.30_2019.csv",
        WeatherDataSourceEnum.NSRDB_15MIN,
    )


@dataclass
class WeatherConfig(ConfigBase):
    """Configuration class for Weather."""

    building_name: str
    name: str
    location: str
    source_path: str
    data_source: WeatherDataSourceEnum
    predictive_control: bool

    @classmethod
    def get_main_classname(cls):
        """Get the name of the main class."""
        return Weather.get_full_classname()

    @classmethod
    def get_default(
        cls,
        location_entry: Any,
        name: str = "Weather",
        building_name: str = "BUI1",
    ) -> Any:
        """Gets the default configuration for Aachen."""
        if not isinstance(location_entry, LocationEnum):
            if location_entry in LocationEnum._member_names_:  # pylint: disable=W0212
                location_entry = getattr(LocationEnum, location_entry)
            else:
                try:
                    location_entry = getattr(LocationEnum, location_entry.strip())
                except Exception as exc:
                    raise ValueError(
                        f"The location_entry {location_entry} could not be found in the Weather Location Enum class."
                    ) from exc

        path = os.path.join(
            utils.get_input_directory(),
            "weather",
            location_entry.value[1],
            location_entry.value[2],
            location_entry.value[3],
        )
        config = WeatherConfig(
            building_name=building_name,
            name=name,
            location=location_entry.value[0],
            source_path=path,
            data_source=location_entry.value[4],
            predictive_control=False,
        )
        return config


class Weather(Component):
    """Provide thermal and solar conditions of local weather."""

    # Inputs
    # None

    # Outputs
    TemperatureOutside = "TemperatureOutside"
    DirectNormalIrradiance = "DirectNormalIrradiance"
    DiffuseHorizontalIrradiance = "DiffuseHorizontalIrradiance"
    DirectNormalIrradianceExtra = "DirectNormalIrradianceExtra"
    GlobalHorizontalIrradiance = "GlobalHorizontalIrradiance"
    Altitude = "Altitude"
    Azimuth = "Azimuth"
    ApparentZenith = "ApparentZenith"
    WindSpeed = "WindSpeed"
    Pressure = "Pressure"
    Weather_Temperature_Forecast_24h = "Weather_Temperature_Forecast_24h"
    DailyAverageOutsideTemperatures = "DailyAverageOutsideTemperatures"

    # Weather_TemperatureOutside_yearly_forecast = "Weather_TemperatureOutside_yearly_forecast"
    # Weather_DiffuseHorizontalIrradiance_yearly_forecast = "Weather_DiffuseHorizontalIrradiance_yearly_forecast"
    # Weather_DirectNormalIrradiance_yearly_forecast = "Weather_DirectNormalIrradiance_yearly_forecast"
    # Weather_DirectNormalIrradianceExtra_yearly_forecast = "Weather_DirectNormalIrradianceExtra_yearly_forecast"
    # Weather_GlobalHorizontalIrradiance_yearly_forecast = "Weather_GlobalHorizontalIrradiance_yearly_forecast"
    # Weather_Azimuth_yearly_forecast = "Weather_Azimuth_yearly_forecast"
    # Weather_ApparentZenith_yearly_forecast = "Weather_ApparentZenith_yearly_forecast"
    Weather_WindSpeed_yearly_forecast = "Weather_WindSpeed_yearly_forecast"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: WeatherConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Initializes the entire class."""
        if my_simulation_parameters is None:
            raise Exception("Simparameters was none")
        self.last_timestep_with_update = -1
        self.weather_config = config
        SingletonSimRepository().set_entry(key=SingletonDictKeyEnum.LOCATION, entry=self.weather_config.location)
        self.parameter_string = my_simulation_parameters.get_unique_key()

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.air_temperature_output: ComponentOutput = self.add_output(
            self.component_name,
            self.TemperatureOutside,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.TemperatureOutside} will follow.",
        )

        self.dni_output: ComponentOutput = self.add_output(
            self.component_name,
            self.DirectNormalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            output_description=f"here a description for {self.DirectNormalIrradiance} will follow.",
        )

        self.dni_extra_output: ComponentOutput = self.add_output(
            self.component_name,
            self.DirectNormalIrradianceExtra,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            output_description=f"here a description for {self.DirectNormalIrradianceExtra} will follow.",
        )

        self.dhi_output: ComponentOutput = self.add_output(
            self.component_name,
            self.DiffuseHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            output_description=f"here a description for {self.DiffuseHorizontalIrradiance} will follow.",
        )

        self.ghi_output: ComponentOutput = self.add_output(
            self.component_name,
            self.GlobalHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            output_description=f"here a description for {self.GlobalHorizontalIrradiance} will follow.",
        )

        self.altitude_output: ComponentOutput = self.add_output(
            self.component_name,
            self.Altitude,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            output_description=f"here a description for {self.Altitude} will follow.",
        )

        self.azimuth_output: ComponentOutput = self.add_output(
            self.component_name,
            self.Azimuth,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            output_description=f"here a description for {self.Azimuth} will follow.",
        )

        self.apparent_zenith_output: ComponentOutput = self.add_output(
            self.component_name,
            self.ApparentZenith,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            output_description=f"here a description for {self.ApparentZenith} will follow.",
        )

        self.wind_speed_output: ComponentOutput = self.add_output(
            self.component_name,
            self.WindSpeed,
            lt.LoadTypes.SPEED,
            lt.Units.METER_PER_SECOND,
            output_description=f"here a description for {self.WindSpeed} will follow.",
        )

        self.pressure_output: ComponentOutput = self.add_output(
            self.component_name,
            self.Pressure,
            lt.LoadTypes.PRESSURE,
            lt.Units.PASCAL,
            output_description=f"here a description for {self.Pressure} will follow.",
        )

        self.daily_average_outside_temperature_output: ComponentOutput = self.add_output(
            self.component_name,
            self.DailyAverageOutsideTemperatures,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.DailyAverageOutsideTemperatures} will follow.",
        )

        self.temperature_list: List[float]
        self.dni_list: List[float]
        self.dniextra_list: List[float]
        self.altitude_list: List[float]
        self.azimuth_list: List[float]
        self.wind_speed_list: List[float]
        self.pressure_list: List[float]
        self.ghi_list: List[float]
        self.apparent_zenith_list: List[float]
        self.dhi_list: List[float]
        self.dry_bulb_list: List[float]
        self.daily_average_outside_temperature_list_in_celsius: List[float]

    def write_to_report(self):
        """Write configuration to the report."""
        return self.weather_config.get_string_dict()

    def i_save_state(self) -> None:
        """Saves the current state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the previous state. Not needed for weather."""
        pass

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Double chekc."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Performs the simulation."""
        if self.last_timestep_with_update == timestep:
            return
        if force_convergence:
            return
        """ Performs the simulation. """
        stsv.set_output_value(self.air_temperature_output, self.temperature_list[timestep])
        stsv.set_output_value(self.dni_output, self.dni_list[timestep])
        stsv.set_output_value(self.dni_extra_output, self.dniextra_list[timestep])
        stsv.set_output_value(self.dhi_output, self.dhi_list[timestep])
        stsv.set_output_value(self.ghi_output, self.ghi_list[timestep])
        stsv.set_output_value(self.altitude_output, self.altitude_list[timestep])
        stsv.set_output_value(self.azimuth_output, self.azimuth_list[timestep])
        stsv.set_output_value(self.wind_speed_output, self.wind_speed_list[timestep])
        stsv.set_output_value(self.apparent_zenith_output, self.apparent_zenith_list[timestep])
        stsv.set_output_value(
            self.pressure_output, self.pressure_list[timestep] * 100
        )  # *100 umrechnung von hPA bzw mbar in PA
        stsv.set_output_value(self.apparent_zenith_output, self.apparent_zenith_list[timestep])
        stsv.set_output_value(
            self.daily_average_outside_temperature_output,
            self.daily_average_outside_temperature_list_in_celsius[timestep],
        )

        # set the temperature forecast
        if self.weather_config.predictive_control:
            timesteps_24h = 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            last_forecast_timestep = int(timestep + timesteps_24h)
            if last_forecast_timestep > len(self.temperature_list):
                last_forecast_timestep = len(self.temperature_list)
            # log.information( type(self.temperature))
            temperatureforecast = self.temperature_list[timestep:last_forecast_timestep]
            self.simulation_repository.set_entry(self.Weather_Temperature_Forecast_24h, temperatureforecast)
        self.last_timestep_with_update = timestep

    def i_prepare_simulation(self) -> None:
        """Generates the lists to be used later."""
        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep
        log.information("Weather config: " + self.weather_config.to_json())  # type: ignore
        location_dict = get_coordinates(
            filepath=self.weather_config.source_path,
            source_enum=self.weather_config.data_source,
        )
        self.simulation_repository.set_entry("weather_location", location_dict)
        cachefound, cache_filepath = utils.get_cache_file("Weather", self.weather_config, self.my_simulation_parameters)
        if cachefound:
            # read cached files
            my_weather = pd.read_csv(cache_filepath, sep=",", decimal=".", encoding="cp1252")
            self.temperature_list = my_weather["t_out"].tolist()
            self.daily_average_outside_temperature_list_in_celsius = my_weather["t_out_daily_average"].tolist()
            self.dry_bulb_list = self.temperature_list
            self.dhi_list = my_weather["DHI"].tolist()
            self.dni_list = my_weather["DNI"].tolist()  # self np.float64( maybe not needed? - Noah
            self.dniextra_list = my_weather["DNIextra"].tolist()
            self.ghi_list = my_weather["GHI"].tolist()
            self.altitude_list = my_weather["altitude"].tolist()
            self.azimuth_list = my_weather["azimuth"].tolist()
            self.apparent_zenith_list = my_weather["apparent_zenith"].tolist()
            self.wind_speed_list = my_weather["Wspd"].tolist()
            try:
                self.pressure_list = my_weather["Pressure"].tolist()
            except Exception:
                log.information("Weather key Pressure could not be found.")
                self.pressure_list = [0] * len(self.wind_speed_list)
        else:
            tmy_data = read_test_reference_year_data(
                weatherconfig=self.weather_config,
                year=self.my_simulation_parameters.year,
            )
            if self.weather_config.data_source == WeatherDataSourceEnum.NSRDB_15MIN:
                dni = tmy_data["DNI"].resample("1T").asfreq().interpolate(method="linear")
                temperature = tmy_data["T"].resample("1T").asfreq().interpolate(method="linear")
                dhi = tmy_data["DHI"].resample("1T").asfreq().interpolate(method="linear")
                ghi = tmy_data["GHI"].resample("1T").asfreq().interpolate(method="linear")
                wind_speed = tmy_data["Wspd"].resample("1T").asfreq().interpolate(method="linear")
                pressure = tmy_data["Pressure"].resample("1T").asfreq().interpolate(method="linear")
            elif self.weather_config.data_source == WeatherDataSourceEnum.DWD_10MIN:
                dni = tmy_data["DNI"].resample("1T").asfreq().interpolate(method="linear")
                temperature = tmy_data["T"].resample("1T").asfreq().interpolate(method="linear")
                dhi = tmy_data["DHI"].resample("1T").asfreq().interpolate(method="linear")
                ghi = tmy_data["GHI"].resample("1T").asfreq().interpolate(method="linear")
                wind_speed = tmy_data["Wspd"].resample("1T").asfreq().interpolate(method="linear")
                pressure = tmy_data["Pressure"].resample("1T").asfreq().interpolate(method="linear")
            elif self.weather_config.data_source == WeatherDataSourceEnum.ERA5:
                dni = tmy_data["DNI"].resample("1T").asfreq().interpolate(method="linear")
                temperature = tmy_data["T"].resample("1T").asfreq().interpolate(method="linear")
                dhi = tmy_data["DHI"].resample("1T").asfreq().interpolate(method="linear")
                ghi = tmy_data["GHI"].resample("1T").asfreq().interpolate(method="linear")
                wind_speed = tmy_data["Wspd"].resample("1T").asfreq().interpolate(method="linear")
                pressure = tmy_data["Pressure"].resample("1T").asfreq().interpolate(method="linear")
            else:
                dni = self.interpolate(tmy_data["DNI"], self.my_simulation_parameters.year)
                temperature = self.interpolate(tmy_data["T"], self.my_simulation_parameters.year)
                dhi = self.interpolate(tmy_data["DHI"], self.my_simulation_parameters.year)
                ghi = self.interpolate(tmy_data["GHI"], self.my_simulation_parameters.year)
                wind_speed = self.interpolate(tmy_data["Wspd"], self.my_simulation_parameters.year)
                pressure = self.interpolate(tmy_data["Pressure"], self.my_simulation_parameters.year)
            # calculate extra terrestrial radiation- n eeded for perez array diffuse irradiance models
            dni_extra = pd.Series(pvlib.irradiance.get_extra_radiation(dni.index), index=dni.index)  # type: ignore

            solpos = pvlib.solarposition.get_solarposition(dni.index, location_dict["latitude"], location_dict["longitude"])  # type: ignore
            altitude = solpos["elevation"]
            azimuth = solpos["azimuth"]
            apparent_zenith = solpos["apparent_zenith"]

            if seconds_per_timestep != 60:
                self.temperature_list = temperature.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.dry_bulb_list = temperature.resample(str(seconds_per_timestep) + "S").mean().to_list()
                self.calculate_daily_average_outside_temperature(
                    temperaturelist=self.temperature_list,
                    seconds_per_timestep=seconds_per_timestep,
                )

                self.dhi_list = dhi.resample(str(seconds_per_timestep) + "S").mean().tolist()
                # np.float64( ## not sure what this is fore. python float and npfloat 64 are the same.
                self.dni_list = dni.resample(str(seconds_per_timestep) + "S").mean().tolist()  # )  # type: ignore
                self.dniextra_list = dni_extra.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.ghi_list = ghi.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.altitude_list = altitude.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.azimuth_list = azimuth.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.apparent_zenith_list = apparent_zenith.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.wind_speed_list = wind_speed.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.pressure_list = pressure.resample(str(seconds_per_timestep) + "S").mean().tolist()
            else:
                self.temperature_list = temperature.tolist()
                self.dry_bulb_list = temperature.to_list()
                self.calculate_daily_average_outside_temperature(
                    temperaturelist=self.temperature_list,
                    seconds_per_timestep=seconds_per_timestep,
                )
                self.dhi_list = dhi.tolist()
                self.dni_list = dni.tolist()
                self.dniextra_list = dni_extra.tolist()
                self.ghi_list = ghi.tolist()
                self.altitude_list = altitude.tolist()
                self.azimuth_list = azimuth.tolist()
                self.apparent_zenith_list = apparent_zenith.tolist()
                self.wind_speed_list = wind_speed.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.pressure_list = pressure.tolist()

            solardata = [
                self.dni_list,
                self.dhi_list,
                self.ghi_list,
                self.temperature_list,
                self.altitude_list,
                self.azimuth_list,
                self.apparent_zenith_list,
                self.dry_bulb_list,
                self.wind_speed_list,
                self.pressure_list,
                self.dniextra_list,
                self.daily_average_outside_temperature_list_in_celsius,
            ]

            database = pd.DataFrame(
                np.transpose(solardata),
                columns=[
                    "DNI",
                    "DHI",
                    "GHI",
                    "t_out",
                    "altitude",
                    "azimuth",
                    "apparent_zenith",
                    "DryBulb",
                    "Wspd",
                    "Pressure",
                    "DNIextra",
                    "t_out_daily_average",
                ],
            )
            database.to_csv(cache_filepath)

        # write one year forecast to simulation repository for PV processing -> if PV forecasts are needed
        if self.weather_config.predictive_control:
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERTEMPERATUREOUTSIDEYEARLYFORECAST,
                entry=self.temperature_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERDIFFUSEHORIZONTALIRRADIANCEYEARLYFORECAST,
                entry=self.dhi_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEYEARLYFORECAST,
                entry=self.dni_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERDIRECTNORMALIRRADIANCEEXTRAYEARLYFORECAST,
                entry=self.dniextra_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERGLOBALHORIZONTALIRRADIANCEYEARLYFORECAST,
                entry=self.ghi_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERAZIMUTHYEARLYFORECAST,
                entry=self.azimuth_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERAPPARENTZENITHYEARLYFORECAST,
                entry=self.apparent_zenith_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERWINDSPEEDYEARLYFORECAST,
                entry=self.wind_speed_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERPRESSUREYEARLYFORECAST,
                entry=self.pressure_list,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.WEATHERALTITUDEYEARLYFORECAST,
                entry=self.altitude_list,
            )

    def interpolate(self, pd_database: Any, year: int) -> Any:
        """Interpolates a time series."""
        firstday = pd.Series(
            [0.0],
            index=[pd.to_datetime(datetime.datetime(year - 1, 12, 31, 23, 0), utc=True).tz_convert(tz="Europe/Berlin")],
        )
        lastday = pd.Series(
            pd_database.iloc[-1],
            index=[pd.to_datetime(datetime.datetime(year, 12, 31, 22, 59), utc=True).tz_convert(tz="Europe/Berlin")],
        )
        pd_database = pd.concat([pd_database, firstday, lastday])
        pd_database = pd_database.sort_index()
        return pd_database.resample("1T").asfreq().interpolate(method="linear")

    def calc_sun_position(self, latitude_deg, longitude_deg, year, hoy):
        """Calculates the Sun Position for a specific hour and location.

        :param latitude_deg: Geographical Latitude in Degrees
        :type latitude_deg: float
        :param longitude_deg: Geographical Longitude in Degrees
        :type longitude_deg: float
        :param year: year
        :type year: int
        :param hoy: Hour of the year from the start. The first hour of January is 1
        :type hoy: int
        :return: altitude, azimuth: Sun position in altitude and azimuth degrees [degrees]
        :rtype: tuple
        """
        # Convert to Radians
        latitude_rad = math.radians(latitude_deg)
        # longitude_rad = math.radians(longitude_deg)  # Note: this is never used

        # Set the date in UTC based off the hour of year and the year itself
        start_of_year = datetime.datetime(year, 1, 1, 0, 0, 0, 0)
        utc_datetime = start_of_year + datetime.timedelta(hours=hoy)

        # Angular distance of the sun north or south of the earths equator
        # Determine the day of the year.
        day_of_year = utc_datetime.timetuple().tm_yday

        # Calculate the declination angle: The variation due to the earths tilt
        # http://www.pveducation.org/pvcdrom/properties-of-sunlight/declination-angle
        declination_rad = math.radians(23.45 * math.sin((2 * math.pi / 365.0) * (day_of_year - 81)))

        # Normalise the day to 2*pi
        # There is some reason as to why it is 364 and not 365.26
        angle_of_day = (day_of_year - 81) * (2 * math.pi / 364)

        # The deviation between local standard time and true solar time
        equation_of_time = (
            (9.87 * math.sin(2 * angle_of_day)) - (7.53 * math.cos(angle_of_day)) - (1.5 * math.sin(angle_of_day))
        )

        # True Solar Time
        solar_time = ((utc_datetime.hour * 60) + utc_datetime.minute + (4 * longitude_deg) + equation_of_time) / 60.0

        # Angle between the local longitude and longitude where the sun is at
        # higher altitude
        hour_angle_rad = math.radians(15 * (12 - solar_time))

        # Altitude Position of the Sun in Radians
        altitude_rad = math.asin(
            math.cos(latitude_rad) * math.cos(declination_rad) * math.cos(hour_angle_rad)
            + math.sin(latitude_rad) * math.sin(declination_rad)
        )

        # Azimuth Position fo the sun in radians
        azimuth_rad = math.asin(math.cos(declination_rad) * math.sin(hour_angle_rad) / math.cos(altitude_rad))

        # I don't really know what this code does, it has been imported from
        # PySolar
        if math.cos(hour_angle_rad) >= (math.tan(declination_rad) / math.tan(latitude_rad)):
            return math.degrees(altitude_rad), math.degrees(azimuth_rad)
        return math.degrees(altitude_rad), (180 - math.degrees(azimuth_rad))

    def calc_sun_position2(self, hoy: Any) -> Any:
        """Calculates the sun position."""
        return self.altitude_list[hoy], self.azimuth_list[hoy]

    def calculate_daily_average_outside_temperature(
        self, temperaturelist: List[float], seconds_per_timestep: int
    ) -> List[float]:
        """Calculate the daily average outside temperatures."""
        timestep_24h = int(24 * 3600 / seconds_per_timestep)
        total_number_of_timesteps_temperature_list = len(temperaturelist)
        self.daily_average_outside_temperature_list_in_celsius = []
        start_index = 0
        for index in range(0, total_number_of_timesteps_temperature_list):
            daily_average_temperature = float(np.mean(temperaturelist[start_index : start_index + timestep_24h]))
            if index == start_index + timestep_24h:
                start_index = index
            self.daily_average_outside_temperature_list_in_celsius.append(daily_average_temperature)
        return self.daily_average_outside_temperature_list_in_celsius

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        opex_cost_data_class = OpexCostDataClass.get_default_opex_cost_data_class()
        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: WeatherConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []


def get_coordinates(filepath: str, source_enum: WeatherDataSourceEnum) -> Any:
    """Reads a test reference year file and gets the GHI, DHI and DNI from it.

    Based on the tsib project @[tsib-kotzur] (Check header)
    """
    # get the correct file path
    # filepath = os.path.join(utils.HISIMPATH["weather"][location])

    if source_enum == WeatherDataSourceEnum.NSRDB_15MIN:
        with open(filepath, encoding="utf-8") as csvfile:
            spamreader = csv.reader(csvfile)
            for i, row in enumerate(spamreader):
                if i == 1:
                    location_name = row[1]
                    lat = float(row[5])
                    lon = float(row[6])
                elif i > 1:
                    break

    elif source_enum == WeatherDataSourceEnum.DWD_10MIN:
        with open(filepath, encoding="utf-8") as csvfile:
            spamreader = csv.reader(csvfile)
            for i, row in enumerate(spamreader):
                if i == 1:
                    location_name = row[0]
                    lat = float(row[1])
                    lon = float(row[2])
                elif i > 1:
                    break

    elif source_enum == WeatherDataSourceEnum.ERA5:
        with open(filepath, encoding="utf-8") as csvfile:
            spamreader = csv.reader(csvfile)
            for i, row in enumerate(spamreader):
                if i == 1:
                    location_name = row[0]
                    lat = float(row[1])
                    lon = float(row[2])
                elif i > 1:
                    break

    else:
        # get the geoposition
        with open(filepath + ".dat", encoding="utf-8") as file_stream:
            lines = file_stream.readlines()
            location_name = lines[0].split(maxsplit=2)[2].replace("\n", "")
            lat = float(lines[1][20:37])
            lon = float(lines[2][15:30])
    return {"name": location_name, "latitude": lat, "longitude": lon}
    # self.index = pd.date_range(f"{year}-01-01 00:00:00", periods=60 * 24 * 365, freq="T", tz="Europe/Berlin")


def read_test_reference_year_data(weatherconfig: WeatherConfig, year: int) -> Any:
    """Reads a test reference year file and gets the GHI, DHI and DNI from it.

    Based on the tsib project @[tsib-kotzur] (Check header)
    """
    # get the correct file path
    filepath = os.path.join(weatherconfig.source_path)
    if weatherconfig.data_source == WeatherDataSourceEnum.NSRDB:
        data = read_nsrdb_data(filepath, year)
    elif weatherconfig.data_source == WeatherDataSourceEnum.DWD_TRY:
        data = read_dwd_try_data(filepath, year)
    elif weatherconfig.data_source == WeatherDataSourceEnum.NSRDB_15MIN:
        data = read_nsrdb_15min_data(filepath, year)
    elif weatherconfig.data_source == WeatherDataSourceEnum.DWD_10MIN:
        data = read_dwd_10min_data(filepath, year)
    elif weatherconfig.data_source == WeatherDataSourceEnum.ERA5:
        data = read_era5_data(filepath, year)

    return data


def read_dwd_try_data(filepath: str, year: int) -> pd.DataFrame:
    """Reads the DWD Test Reference Year (TRY) data."""
    # get the geoposition
    with open(filepath + ".dat", encoding="utf-8") as file_stream:
        lines = file_stream.readlines()
        lat = float(lines[1][20:37])
        lon = float(lines[2][15:30])
    # check if time series data already exists as .csv with DNI
    if os.path.isfile(filepath + ".csv"):
        data = pd.read_csv(filepath + ".csv", index_col=0, parse_dates=True, sep=";", decimal=",")
        data.index = pd.to_datetime(data.index, utc=True).tz_convert("Europe/Berlin")
    # else read from .dat and calculate DNI etc.
    else:
        # get data
        data = pd.read_csv(filepath + ".dat", sep=r"\s+", skiprows=list(range(0, 31)))
        data.index = pd.date_range(f"{year}-01-01 00:30:00", periods=8760, freq="H", tz="Europe/Berlin")
        data["GHI"] = data["D"] + data["B"]
        data = data.rename(
            columns={
                "D": "DHI",
                "t": "T",
                "WG": "Wspd",
                "MM": "Month",
                "DD": "Day",
                "HH": "Hour",
                "p": "Pressure",
                "WR": "Wdir",
            }
        )

        # calculate direct normal
        data["DNI"] = calculate_direct_normal_radiation(data["B"], lon, lat)
    return data


def read_nsrdb_data(filepath: str, year: int) -> pd.DataFrame:
    """Reads a set of NSRDB data."""
    # get data
    data = pd.read_csv(filepath + ".dat", sep=",", skiprows=list(range(0, 11)))
    data = data.drop(data.index[8761:8772])
    data.index = pd.date_range(f"{year}-01-01 00:30:00", periods=8760, freq="H", tz="Europe/Berlin")
    data = data.rename(
        columns={
            "DHI": "DHI",
            "Temperature": "T",
            "Wind Speed": "Wspd",
            "MM": "Month",
            "DD": "Day",
            "HH": "Hour",
            "Pressure": "Pressure",
            "Wind Direction": "Wdir",
            "GHI": "GHI",
            "DNI": "DNI",
        }
    )
    return data


def read_nsrdb_15min_data(filepath: str, year: int) -> pd.DataFrame:
    """Reads a set of NSRDB data in 15 min resolution."""
    data = pd.read_csv(filepath, encoding="utf-8", skiprows=[0, 1])
    # get data
    data.index = pd.date_range(f"{year}-01-01 00:00:00", periods=24 * 4 * 365, freq="900S", tz="UTC")
    data = data.rename(
        columns={
            "Temperature": "T",
            "Wind Speed": "Wspd",
        }
    )
    return data


def read_dwd_10min_data(filepath: str, year: int) -> pd.DataFrame:
    """Reads a set of DWD data in 10 min resolution.

    https://github.com/earthobservations/wetterdienst/tree/main
    """

    # get location
    location = pd.read_csv(  # type: ignore
        filepath,
        nrows=1,
        skiprows=1,
        header=None,
        names=pd.read_csv(filepath, nrows=1).columns,
    )
    longitude = location["longitude"][0]
    latitude = location["latitude"][0]

    # get data
    data = pd.read_csv(filepath, encoding="utf-8", skiprows=[0, 1])
    data.index = pd.date_range(f"{year}-01-01 00:00:00", periods=24 * 6 * 365, freq="600S", tz="UTC")
    data = data.rename(
        columns={
            "diffuse_irradiance": "DHI",
            "temperature": "T",
            "wind_speed": "Wspd",
            "month": "Month",
            "day": "Day",
            "hour": "Hour",
            "minute": "Minutes",
            "pressure": "Pressure",
            "wind_direction": "Wdir",
            "global_irradiance": "GHI",
        }
    )
    # calculate direct normal
    data["direct_horizontal_irradiance"] = data["GHI"] - data["DHI"]
    data["DNI"] = calculate_direct_normal_radiation(data["direct_horizontal_irradiance"], longitude, latitude)

    return data


def read_era5_data(filepath: str, year: int) -> pd.DataFrame:
    """Reads a set of era5 in 60 min resolution.

    https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview
    """

    # get location
    location = pd.read_csv(  # type: ignore
        filepath,
        nrows=1,
        skiprows=1,
        header=None,
        names=pd.read_csv(filepath, nrows=1).columns,
    )
    longitude = location["longitude"][0]
    latitude = location["latitude"][0]

    # get data
    data = pd.read_csv(filepath, encoding="utf-8", skiprows=[0, 1])
    data.index = pd.date_range(f"{year}-01-01 00:00:00", periods=8760, freq="H", tz="UTC")
    data = data.rename(
        columns={
            "month": "Month",
            "day": "Day",
            "hour": "Hour",
            "minute": "Minutes",
            "temperature": "T",
            "pressure": "Pressure",
            "wind_direction": "Wdir",
            "wind_speed": "Wspd",
            "global_irradiance": "GHI",
        }
    )
    # calculate direct normal
    data["DHI"] = data["GHI"] - data["direct_irradiance"]
    data["DNI"] = calculate_direct_normal_radiation(data["direct_irradiance"], longitude, latitude)

    return data


def calculate_direct_normal_radiation(
    direct_horizontal_irradation: pd.Series,
    lon: float,
    lat: float,
    zenith_tol: float = 87.0,
) -> pd.Series:
    """Calculates the direct NORMAL irradiance from the direct horizontal irradiance with the help of the PV lib.

    Based on the tsib project @[tsib-kotzur] (Check header)

    Parameters
    ----------
    direct_horizontal_irradation: pd.Series with time index
        Direct horizontal irradiance
    lon: float
        Longitude of the location
    lat: float
        Latitude of the location
    zenith_tol: float, optional
        Avoid cosines of values above a certain zenith angle of in order to avoid division by zero.

    Returns
    -------
    dni: pd.Series

    """

    solar_pos = pvlib.solarposition.get_solarposition(direct_horizontal_irradation.index, lat, lon)
    solar_pos["apparent_zenith"][solar_pos.apparent_zenith > zenith_tol] = zenith_tol
    dni = direct_horizontal_irradation.div(solar_pos["apparent_zenith"].apply(math.radians).apply(math.cos))
    if sum(dni.isnull()) > 0:
        raise ValueError("Something went wrong...")
    return dni



"""Example Transformer."""

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
from typing import List, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Import modules from HiSim
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput, DisplayConfig
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim.component import ConfigBase


@dataclass_json
@dataclass
class ExampleTransformerConfig(ConfigBase):
    """Configuration of the Example Transformer."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return ExampleTransformer.get_full_classname()

    building_name: str
    # parameter_string: str
    # my_simulation_parameters: SimulationParameters
    name: str
    loadtype: lt.LoadTypes
    unit: lt.Units

    @classmethod
    def get_default_transformer(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default Transformer."""
        return ExampleTransformerConfig(
            building_name=building_name,
            name="Example Transformer default",
            loadtype=lt.LoadTypes.ANY,
            unit=lt.Units.ANY,
        )


class ExampleTransformer(Component):
    """The Example Transformer class.

    It is used to modify input values and return them as new output values.

    Parameters
    ----------
    component_name : str
        Passed to initialize :py:class:`~hisim.component.Component`.

    loadtype : LoadType
        A :py:class:`~hisim.loadtypes.LoadTypes` object that represents
        the type of the loaded data.

    unit: LoadTypes.Units
        A :py:class:`~hisim.loadtypes.Units` object that represents
        the unit of the loaded data.

    """

    TransformerInput = "Input1"
    TransformerInput2 = "Optional Input1"
    TransformerOutput = "MyTransformerOutput"
    TransformerOutput2 = "MyTransformerOutput2"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ExampleTransformerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Constructs all the neccessary attributes."""
        self.transformerconfig = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.input1: ComponentInput = self.add_input(
            self.transformerconfig.name,
            ExampleTransformer.TransformerInput,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )
        self.input2: ComponentInput = self.add_input(
            self.transformerconfig.name,
            ExampleTransformer.TransformerInput2,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            False,
        )
        self.output1: ComponentOutput = self.add_output(
            self.transformerconfig.name,
            ExampleTransformer.TransformerOutput,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Output 1",
        )
        self.output2: ComponentOutput = self.add_output(
            self.transformerconfig.name,
            ExampleTransformer.TransformerOutput2,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Output 2",
        )

    def i_save_state(self) -> None:
        """Saves the current state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_restore_state(self) -> None:
        """Restores previous state."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the transformer."""
        startval_1 = stsv.get_input_value(self.input1)
        startval_2 = stsv.get_input_value(self.input2)
        stsv.set_output_value(self.output1, startval_1 * 5)
        stsv.set_output_value(self.output2, startval_2 * 1000)

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        lines = []
        lines.append("Transformer: " + self.component_name)
        return lines
